library wlx_fb2;

uses
  SysUtils,
  Windows,
  Classes,
  XMLIntf,
  XMLDoc,
  Shared,
  XMLShared,
  UConvert,
  System.Zip,
  System.Types,
  MobiUnit in 'MobiUnit.pas';


{$IFDEF WIN64}
  {$E w64}
  {$IFNDEF VER230}
    {$E wlx64}
  {$ENDIF}
{$ENDIF}
{$IFDEF WIN32}
  {$E wlx}
{$ENDIF}

{$R *.res}
{$WEAKLINKRTTI ON}
{$SETPEOPTFLAGS $160}

const
  PARSE_FUNCTION = 'EXT="FB2"|EXT="FBZ"|EXT="EPUB"|EXT="MOBI"|EXT="AZW"|EXT="AZW3"';

var
  RestrictToFB2ZIP: Boolean = False;
  MobiAllowFirstPic: Boolean = False;


procedure ListGetDetectString(DetectString: PAnsiChar;
  MaxLen: integer); stdcall;
begin
  StrLCopy(DetectString, PAnsiChar(PARSE_FUNCTION), MaxLen - 1);
end;

function ListLoadW(ParentWin: thandle; FileToLoad: pwidechar;
  ShowFlags: integer): thandle; stdcall;
begin
  Result := 0;
end;

function ListLoad(ParentWin: thandle; FileToLoad: PAnsiChar;
  ShowFlags: integer): thandle; stdcall;
begin
  Result := 0;
end;

function IniReadString(const FileName, Section, Ident, Default: string): string;
var
  Buffer: array[0..2047] of Char;
begin
  SetString(Result, Buffer, GetPrivateProfileString(PChar(Section),
    PChar(Ident), PChar(Default), Buffer, Length(Buffer), PChar(FileName)));
end;

function ExtractFileNameEx(const FileName: string): string;
var
  I: Integer;
begin
  I := LastDelimiter('/', FileName);
  Result := Copy(FileName, I + 1, MaxInt);
end;

function ExtractFileDirEx(const FileName: string): string;
var
  I: Integer;
begin
  I := LastDelimiter('/', Filename);
  if (I > 1) and (FileName[I] = '/') and
    (not IsDelimiter('/', FileName, I-1)) then Dec(I);
  Result := Copy(FileName, 1, I);
end;

function ExtractFilePathEx(const FileName: string): string;
var
  I: Integer;
begin
  I := LastDelimiter('/', FileName);
  Result := Copy(FileName, 1, I);
end;

function ListGetPreviewBitmapW(FileToLoad: pwidechar; width, height: integer;
  contentbuf: PAnsiChar; contentbuflen: integer): hbitmap; stdcall;
var
  FileBuf, imgbuf, imgname, ArcFile, FileName, imgpath: string;
  raw: AnsiString;
  i: integer;
  xmlbook: IXMLDocument;
  xmlbookRoot, t, desc: IXMLNode;
  MS: TMemoryStream;
  ZipFile: TZipFile;
  Flag: Boolean;
  ZH: TZipHeader;
  Stream: TStream;
  FS: TFileStream;
  PS: TPointerReadOnlyStream;
  B: TBytes;
begin
  Result := 0;
  Flag := False;
  Stream := nil;
  xmlbook := nil;
  FileName := GetVeryLongFileName(FileToLoad);

  try
    FileBuf := ExtractFileExt(FileToLoad);
    FileBuf := AnsiLowerCase(FileBuf);

    if (FileBuf = '.mobi') or (FileBuf = '.azw') or (FileBuf = '.azw3') or
      (FileBuf = '.azw4') or (FileBuf = '.kf8') then
    begin
      B := GetMobiCover(FileName, MobiAllowFirstPic);
      if Length(B) = 0 then
        Exit;

      PS := TPointerReadOnlyStream.Create(Pointer(B), Length(B));
      try
        Result := ConvertStream(PS, Point(width, height));
      finally
        PS.Free;
      end;

      Exit; // mobi
    end;

    if (FileBuf = '.fbz') or (FileBuf = '.fb2z') or (FileBuf = '.zfb2') or
      (FileBuf = '.zip') or (FileBuf = '.epub') then
    begin
      if RestrictToFB2ZIP and (FileBuf = '.zip') then
        if not SameText(Copy(FileName, Length(FileName) - 6, MaxInt), 'fb2.zip') then
          Exit;

      if not TZipFile.IsValid(FileName) then
        Exit;

      ZipFile := TZipFile.Create;
      try

        ZipFile.Open(FileName, zmRead);

        if FileBuf = '.epub' then
        begin
          { Sometimes TZipFile return in stream opened zip file... Use TBytes }
          ZipFile.Read('META-INF/container.xml', B);
          PS := TPointerReadOnlyStream.Create(Pointer(B), Length(B));
          try
            xmlbook := TXMLDocument.Create('');
            xmlbook.XML.LoadFromStream(PS);
          finally
            PS.Free;
          end;
          if xmlbook = nil then
            Exit;
          xmlbook.Options := [];
          xmlbookRoot := GetRootNode(xmlbook, 'container');
          if xmlbookRoot = nil then
            Exit;
          t := GetNodeByLongName(xmlbookRoot, 'rootfiles/rootfile');
          if t = nil then
            Exit;
          t := t.AttributeNodes.FindNode('full-path');
          if t = nil then
            Exit;

          FreeAndNil(Stream);
          { Sometimes TZipFile return in stream opened zip file... Use TBytes }
          ZipFile.Read(t.Text, B);
          PS := TPointerReadOnlyStream.Create(Pointer(B), Length(B));
          try
            imgpath := ExtractFilePathEx(t.Text);
            xmlbook.LoadFromStream(PS);
          finally
            PS.Free;
          end;

          xmlbook.Options := [];
          xmlbookRoot := GetRootNode(xmlbook, 'package');
          if xmlbookRoot = nil then
            Exit;
          t := GetNodeByLongName(xmlbookRoot, 'metadata');
          if t = nil then
            Exit;

          imgbuf := '';
          for I := 0 to t.ChildNodes.Count - 1 do
          begin
            if t.ChildNodes.Nodes[I].NodeName <> 'meta' then
              Continue;
            if SameText(t.ChildNodes[I].Attributes['name'], 'cover') then
              imgbuf := t.ChildNodes[I].Attributes['content'];
            if imgbuf = '' then
              Continue;
          end;
          if imgbuf = '' then
            Exit;

          imgname := '';
          t := GetNodeByLongName(xmlbookRoot, 'manifest');
          if t = nil then
            Exit;
          for I := 0 to t.ChildNodes.Count - 1 do
          begin
            if t.ChildNodes[I].Attributes['id'] = imgbuf then
              imgname := t.ChildNodes[I].Attributes['href'];
            if imgname = '' then
              Continue;
          end;
          if imgname = '' then
            Exit;

          t := nil;
          desc := nil;
          xmlbookRoot := nil;
          xmlbook := nil;

          FreeAndNil(Stream);
          if Pos('../', imgname) > 0 then
            imgname := StringReplace(imgname, '../', '', [rfReplaceAll]) // simple try root
          else
            imgname := imgpath + imgname;
          { Sometimes TZipFile return in stream opened zip file... Use TBytes }
          ZipFile.Read(imgname, B);
          PS := TPointerReadOnlyStream.Create(Pointer(B), Length(B));
          try
            Result := ConvertStream(PS, Point(width, height));
          finally
            PS.Free;
          end;

          Exit; // epub
        end;


        for ArcFile in ZipFile.FileNames do
          if SameText(ExtractFileExt(ArcFile), '.fb2') then
          begin
            Flag := True;
            Break;
          end;

        if not Flag then
          Exit;

        ZipFile.Read(ArcFile, Stream, ZH);

        xmlbook := TXMLDocument.Create('');
        xmlbook.XML.LoadFromStream(Stream);

      finally
        FreeAndNil(Stream);
        ZipFile.Free;
      end;

    end
    else
      { Very long file names can't be handled directly by TXMLDocument }
      if Pos('\\?\', FileName) > 0 then
      begin
        FS := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
        try
          xmlbook := TXMLDocument.Create('');
          xmlbook.XML.LoadFromStream(FS);
        finally
          FS.Free;
        end;
      end else
        xmlbook := TXMLDocument.Create(FileName);

    if xmlbook = nil then
      Exit;
    xmlbook.Options := [];
    xmlbookRoot := GetRootNode(xmlbook, 'FictionBook');
    if xmlbookRoot = nil then
      Exit;
    {
    t := GetNodeByLongName(xmlbookRoot,
      'description/title-info/coverpage/image');
    if t = nil then
      t := GetNodeByLongName(xmlbookRoot,
        'description/src-title-info/coverpage/image');
    if t = nil then
      Exit;
    }

    desc := xmlbookRoot.ChildNodes.FindNode('description');
    if desc = nil then
      Exit;
    t := GetNodeByLongName(desc,
      'title-info/coverpage/image');
    if t = nil then
      t := GetNodeByLongName(desc,
        'src-title-info/coverpage/image');
    if t = nil then
      Exit;

    imgname := t.AttributeNodes[0].Text;
    if imgname = '' then
      Exit;
    if imgname[1] = '#' then
      imgname := copy(imgname, 2, maxint);
    imgbuf := '';
    for i := 0 to xmlbookRoot.ChildNodes.Count - 1 do
    begin
      t := xmlbookRoot.ChildNodes[i];
      if (t.NodeName = 'binary') and (t.Attributes['id'] = imgname) then
      begin
        imgbuf := t.Text;
        Break;
      end;
    end;
    if imgbuf = '' then
      Exit;

    t := nil;
    desc := nil;
    xmlbookRoot := nil;
    xmlbook := nil;

    //raw := DecodeBase64(AnsiString(imgbuf));
    raw := DecodeBase64U(imgbuf);

    if raw = '' then
      Exit;

    {
    MS := TMemoryStream.Create;
    try
      //MS.WriteBuffer(raw[1], Length(raw));
      MS.WriteBuffer(PAnsiChar(raw)^, Length(raw));
      MS.Position := 0;
      Result := ConvertStream(MS, Point(width, height));
    finally
      MS.Free;
    end;
    }
    PS := TPointerReadOnlyStream.Create(Pointer(raw), Length(raw));
    try
      Result := ConvertStream(PS, Point(width, height));
    finally
      PS.Free;
    end;
  except
    //on E: Exception do
      //MessageBox(GetActiveWindow, PChar(E.Message + #13 + FileName), nil, 0);
  end;
end;

function ListGetPreviewBitmap(FileToLoad: PAnsiChar; width, height: integer;
  contentbuf: PAnsiChar; contentbuflen: integer): hbitmap; stdcall;
begin
  Result := ListGetPreviewBitmapW(pwidechar(WideString(AnsiString(FileToLoad))),
    width, height, contentbuf, contentbuflen);
end;

exports
  ListGetDetectString,
  ListLoad,
  ListLoadW,
  ListGetPreviewBitmapW,
  ListGetPreviewBitmap;

begin
  //ReportMemoryLeaksOnShutdown := True;

  RestrictToFB2ZIP := IniReadString(ChangeFileExt(GetModuleName(HInstance), '.ini'),
    'wlx_fb2', 'RestrictToFB2ZIP', '0') = '1';
  MobiAllowFirstPic := IniReadString(ChangeFileExt(GetModuleName(HInstance), '.ini'),
    'wlx_fb2', 'MobiAllowFirstPic', '0') = '1';

end.
